/*
 * Copyright (C) 1997-2010 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file command.c Client authentication and server commands. */

#include <regex.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <dirent.h>
#include <openssl/rc4.h>
#include <osl.h>

#include "para.h"
#include "error.h"
#include "crypt.h"
#include "command.h"
#include "server.cmdline.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "vss.h"
#include "list.h"
#include "send.h"
#include "rc4.h"
#include "net.h"
#include "daemon.h"
#include "fd.h"
#include "ipc.h"
#include "user_list.h"
#include "server_command_list.h"
#include "afs_command_list.h"
#include "sched.h"
#include "signal.h"

/** Commands including options must be shorter than this. */
#define MAX_COMMAND_LEN 32768

extern int mmd_mutex;
extern struct misc_meta_data *mmd;
extern struct sender senders[];
int send_afs_status(struct rc4_context *rc4c, int parser_friendly);

const char *status_item_list[] = {STATUS_ITEM_ARRAY};

static void dummy(__a_unused int s)
{
}

static void mmd_dup(struct misc_meta_data *new_mmd)
{
	mutex_lock(mmd_mutex);
	*new_mmd = *mmd;
	mutex_unlock(mmd_mutex);
}

/*
 * Compute human readable string containing vss status for given integer value.
 *
 * We don't want to use vss_playing() and friends here because we take a
 * snapshot of the mmd struct and use the copy for computing the state of the
 * vss. If the real data were used, we would take the mmd lock for a rather
 * long time or risk to get an inconsistent view.
 */
static char *vss_status_tohuman(unsigned int flags)
{
	if (flags & VSS_PLAYING)
		return para_strdup("playing");
	if (flags & VSS_NEXT)
		return para_strdup("stopped");
	return para_strdup("paused");
}

/*
 * return human readable permission string. Never returns NULL.
 */
static char *cmd_perms_itohuman(unsigned int perms)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = perms & AFS_READ? 'a' : '-';
	msg[1] = perms & AFS_WRITE? 'A' : '-';
	msg[2] = perms & VSS_READ? 'v' : '-';
	msg[3] = perms & VSS_WRITE? 'V' : '-';
	msg[4] = '\0';
	return msg;
}

/*
 * Never returns NULL.
 */
static char *vss_get_status_flags(unsigned int flags)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = (flags & VSS_PLAYING)? 'P' : '_';
	msg[1] = (flags & VSS_NOMORE)? 'O' : '_';
	msg[2] = (flags & VSS_NEXT)? 'N' : '_';
	msg[3] = (flags & VSS_REPOS)? 'R' : '_';
	msg[4] = '\0';
	return msg;
}

static char *get_status(struct misc_meta_data *nmmd, int parser_friendly)
{
	char mtime[30] = "";
	char *status, *flags; /* vss status info */
	char *ut = uptime_str();
	long offset = (nmmd->offset + 500) / 1000;
	struct timeval current_time;
	struct tm mtime_tm;
	struct para_buffer b = {.flags = parser_friendly? PBF_SIZE_PREFIX : 0};

	/* report real status */
	status = vss_status_tohuman(nmmd->vss_status_flags);
	flags = vss_get_status_flags(nmmd->vss_status_flags);
	if (nmmd->size) { /* parent currently has an audio file open */
		localtime_r(&nmmd->mtime, &mtime_tm);
		strftime(mtime, 29, "%b %d %Y", &mtime_tm);
	}
	gettimeofday(&current_time, NULL);
	WRITE_STATUS_ITEM(&b, SI_FILE_SIZE, "%zu\n", nmmd->size / 1024);
	WRITE_STATUS_ITEM(&b, SI_MTIME, "%s\n", mtime);
	WRITE_STATUS_ITEM(&b, SI_STATUS, "%s\n", status);
	WRITE_STATUS_ITEM(&b, SI_STATUS_FLAGS, "%s\n", flags);
	WRITE_STATUS_ITEM(&b, SI_OFFSET, "%li\n", offset);
	WRITE_STATUS_ITEM(&b, SI_AFS_MODE, "%s\n", mmd->afs_mode_string);
	WRITE_STATUS_ITEM(&b, SI_STREAM_START, "%lu.%lu\n",
		(long unsigned)nmmd->stream_start.tv_sec,
		(long unsigned)nmmd->stream_start.tv_usec);
	WRITE_STATUS_ITEM(&b, SI_CURRENT_TIME, "%lu.%lu\n",
		(long unsigned)current_time.tv_sec,
		(long unsigned)current_time.tv_usec);
	free(flags);
	free(status);
	free(ut);
	return b.buf;
}

static int check_sender_args(int argc, char * const * argv, struct sender_command_data *scd)
{
	int i;
	/* this has to match sender.h */
	const char *subcmds[] = {"add", "delete", "allow", "deny", "on", "off", NULL};

	scd->sender_num = -1;
	if (argc < 2)
		return -E_COMMAND_SYNTAX;
	for (i = 0; senders[i].name; i++)
		if (!strcmp(senders[i].name, argv[1]))
			break;
	PARA_DEBUG_LOG("%d:%s\n", argc, argv[1]);
	if (!senders[i].name)
		return -E_COMMAND_SYNTAX;
	scd->sender_num = i;
	for (i = 0; subcmds[i]; i++)
		if (!strcmp(subcmds[i], argv[2]))
			break;
	if (!subcmds[i])
		return -E_COMMAND_SYNTAX;
	scd->cmd_num = i;
	if (!senders[scd->sender_num].client_cmds[scd->cmd_num])
		return -E_SENDER_CMD;
	switch (scd->cmd_num) {
	case SENDER_ON:
	case SENDER_OFF:
		if (argc != 3)
			return -E_COMMAND_SYNTAX;
		break;
	case SENDER_DENY:
	case SENDER_ALLOW:
		if (argc != 4 || parse_cidr(argv[3], scd->host,
				sizeof(scd->host), &scd->netmask) == NULL)
			return -E_COMMAND_SYNTAX;
		break;
	case SENDER_ADD:
	case SENDER_DELETE:
		if (argc != 4)
			return -E_COMMAND_SYNTAX;
		return parse_fec_url(argv[3], scd);
	default:
		return -E_COMMAND_SYNTAX;
	}
	return 1;
}

int com_sender(struct rc4_context *rc4c, int argc, char * const * argv)
{
	int i, ret;
	struct sender_command_data scd;

	if (argc < 2) {
		char *msg = NULL;
		for (i = 0; senders[i].name; i++) {
			char *tmp = make_message("%s%s\n",
				msg? msg : "", senders[i].name);
			free(msg);
			msg = tmp;
		}
		ret = rc4_send_buffer(rc4c, msg);
		free(msg);
		return ret;
	}
	ret = check_sender_args(argc, argv, &scd);
	if (ret < 0) {
		char *msg;
		if (scd.sender_num < 0)
			return ret;
		msg = senders[scd.sender_num].help();
		ret = rc4_send_buffer(rc4c, msg);
		free(msg);
		return ret;
	}
	for (i = 0; i < 10; i++) {
		mutex_lock(mmd_mutex);
		if (mmd->sender_cmd_data.cmd_num >= 0) {
			mutex_unlock(mmd_mutex);
			usleep(100 * 1000);
			continue;
		}
		memcpy(&mmd->sender_cmd_data, &scd, sizeof(scd));
		mutex_unlock(mmd_mutex);
		break;
	}
	return (i < 10)? 1 : -E_LOCK;
}

/* server info */
int com_si(struct rc4_context *rc4c, int argc, __a_unused char * const * argv)
{
	int i, ret;
	char *ut;
	char *sender_info = NULL;

	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	for (i = 0; senders[i].name; i++) {
		char *info = senders[i].info();
		sender_info = para_strcat(sender_info, info);
		free(info);
	}
	ut = uptime_str();
	ret = rc4_send_va_buffer(rc4c, "up: %s\nplayed: %u\n"
		"server_pid: %d\n"
		"afs_pid: %d\n"
		"connections (active/accepted/total): %u/%u/%u\n"
		"current loglevel: %s\n"
		"supported audio formats: %s\n"
		"%s",
		ut, mmd->num_played,
		(int)getppid(),
		(int)mmd->afs_pid,
		mmd->active_connections,
		mmd->num_commands,
		mmd->num_connects,
		conf.loglevel_arg,
		supported_audio_formats(),
		sender_info
	);
	mutex_unlock(mmd_mutex);
	free(ut);
	free(sender_info);
	return ret;
}

/* version */
int com_version(struct rc4_context *rc4c, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	return rc4_send_buffer(rc4c, VERSION_TEXT("server")
		"built: " BUILD_DATE "\n"
		UNAME_RS ", " CC_VERSION "\n"
	);
}

#define EMPTY_STATUS_ITEMS \
	ITEM(PATH) \
	ITEM(DIRECTORY) \
	ITEM(BASENAME) \
	ITEM(SCORE) \
	ITEM(ATTRIBUTES_BITMAP) \
	ITEM(ATTRIBUTES_TXT) \
	ITEM(HASH) \
	ITEM(IMAGE_ID) \
	ITEM(IMAGE_NAME) \
	ITEM(LYRICS_ID) \
	ITEM(LYRICS_NAME) \
	ITEM(BITRATE) \
	ITEM(FORMAT) \
	ITEM(FREQUENCY) \
	ITEM(CHANNELS) \
	ITEM(DURATION) \
	ITEM(SECONDS_TOTAL) \
	ITEM(NUM_PLAYED) \
	ITEM(LAST_PLAYED) \
	ITEM(TECHINFO) \
	ITEM(ARTIST) \
	ITEM(TITLE) \
	ITEM(YEAR) \
	ITEM(ALBUM) \
	ITEM(COMMENT) \
	ITEM(AMPLIFICATION)

/**
 * Write a list of audio-file related status items with empty values.
 *
 * This is used by vss when currently no audio file is open.
 */
static char *empty_status_items(int parser_friendly)
{
	if (parser_friendly)
		return make_message(
			#define ITEM(x) "0004 %02x:\n"
			EMPTY_STATUS_ITEMS
			#undef ITEM
			#define ITEM(x) , SI_ ## x
			EMPTY_STATUS_ITEMS
			#undef ITEM
		);
	return make_message(
		#define ITEM(x) "%s:\n"
		EMPTY_STATUS_ITEMS
		#undef ITEM
		#define ITEM(x) ,status_item_list[SI_ ## x]
		EMPTY_STATUS_ITEMS
		#undef ITEM
	);
}
#undef EMPTY_STATUS_ITEMS

/* stat */
int com_stat(struct rc4_context *rc4c, int argc, char * const * argv)
{
	int i, ret;
	struct misc_meta_data tmp, *nmmd = &tmp;
	char *s;
	int32_t num = 0;
	int parser_friendly = 0;

	para_sigaction(SIGUSR1, dummy);

	for (i = 1; i < argc; i++) {
		const char *arg = argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strncmp(arg, "-n=", 3)) {
			ret = para_atoi32(arg + 3, &num);
			if (ret < 0)
				return ret;
			continue;
		}
		if (!strcmp(arg, "-p")) {
			parser_friendly = 1;
			continue;
		}
		return -E_COMMAND_SYNTAX;
	}
	if (i != argc)
		return -E_COMMAND_SYNTAX;
	for (;;) {
		mmd_dup(nmmd);
		s = get_status(nmmd, parser_friendly);
		ret = rc4_send_buffer(rc4c, s);
		free(s);
		if (ret < 0)
			goto out;
		if (nmmd->vss_status_flags & VSS_NEXT) {
			static char *esi;
			if (!esi)
				esi = empty_status_items(parser_friendly);
			ret = rc4_send_buffer(rc4c, esi);
			if (ret < 0)
				goto out;
		} else
			send_afs_status(rc4c, parser_friendly);
		ret = 1;
		if (num > 0 && !--num)
			goto out;
		sleep(50);
		if (getppid() == 1)
			return -E_SERVER_CRASH;
	}
out:
	return ret;
}

static int send_list_of_commands(struct rc4_context *rc4c, struct server_command *cmd,
		const char *handler)
{
	int ret, i;

	for (i = 1; cmd->name; cmd++, i++) {
		char *perms = cmd_perms_itohuman(cmd->perms);
		ret = rc4_send_va_buffer(rc4c, "%s\t%s\t%s\t%s\n", cmd->name,
			handler,
			perms,
			cmd->description);
		free(perms);
		if (ret < 0)
			return ret;
	}
	return 1;
}

/* returns string that must be freed by the caller */
static struct server_command *get_cmd_ptr(const char *name, char **handler)
{
	struct server_command *cmd;

	for (cmd = server_cmds; cmd->name; cmd++)
		if (!strcmp(cmd->name, name)) {
			if (handler)
				*handler = para_strdup("server"); /* server commands */
			return cmd;
		}
	/* not found, look for commands supported by afs */
	for (cmd = afs_cmds; cmd->name; cmd++)
		if (!strcmp(cmd->name, name)) {
			if (handler)
				*handler = para_strdup("afs");
			return cmd;
		}
	return NULL;
}

/* help */
int com_help(struct rc4_context *rc4c, int argc, char * const * argv)
{
	struct server_command *cmd;
	char *perms, *handler;
	int ret;

	if (argc < 2) {
		/* no argument given, print list of commands */
		if ((ret = send_list_of_commands(rc4c, server_cmds, "server")) < 0)
			return ret;
		return send_list_of_commands(rc4c, afs_cmds, "afs");
	}
	/* argument given for help */
	cmd = get_cmd_ptr(argv[1], &handler);
	if (!cmd) {
		free(handler);
		return -E_BAD_CMD;
	}
	perms = cmd_perms_itohuman(cmd->perms);
	ret = rc4_send_va_buffer(rc4c,
		"%s - %s\n\n"
		"handler: %s\n"
		"permissions: %s\n"
		"usage: %s\n\n"
		"%s\n",
		argv[1],
		cmd->description,
		handler,
		perms,
		cmd->usage,
		cmd->help
	);
	free(perms);
	free(handler);
	return ret;
}

/* hup */
int com_hup(__a_unused struct rc4_context *rc4c, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	kill(getppid(), SIGHUP);
	return 1;
}

/* term */
int com_term(__a_unused struct rc4_context *rc4c, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	kill(getppid(), SIGTERM);
	return 1;
}

int com_play(__a_unused struct rc4_context *rc4c, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	mmd->new_vss_status_flags |= VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_NOMORE;
	mutex_unlock(mmd_mutex);
	return 1;

}

/* stop */
int com_stop(__a_unused struct rc4_context *rc4c, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	mmd->new_vss_status_flags &= ~VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_REPOS;
	mmd->new_vss_status_flags |= VSS_NEXT;
	mutex_unlock(mmd_mutex);
	return 1;
}

/* pause */
int com_pause(__a_unused struct rc4_context *rc4c, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	if (!vss_paused() && !vss_stopped()) {
		mmd->events++;
		mmd->new_vss_status_flags &= ~VSS_PLAYING;
		mmd->new_vss_status_flags &= ~VSS_NEXT;
	}
	mutex_unlock(mmd_mutex);
	return 1;
}

/* next */
int com_next(__a_unused struct rc4_context *rc4c, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	mmd->events++;
	mmd->new_vss_status_flags |= VSS_NEXT;
	mutex_unlock(mmd_mutex);
	return 1;
}

/* nomore */
int com_nomore(__a_unused struct rc4_context *rc4c, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	if (vss_playing() || vss_paused())
		mmd->new_vss_status_flags |= VSS_NOMORE;
	mutex_unlock(mmd_mutex);
	return 1;
}

/* ff */
int com_ff(__a_unused struct rc4_context *rc4c, int argc, char * const * argv)
{
	long promille;
	int ret, backwards = 0;
	unsigned i;
	char c;

	if (argc != 2)
		return -E_COMMAND_SYNTAX;
	if (!(ret = sscanf(argv[1], "%u%c", &i, &c)))
		return -E_COMMAND_SYNTAX;
	if (ret > 1 && c == '-')
		backwards = 1; /* jmp backwards */
	mutex_lock(mmd_mutex);
	ret = -E_NO_AUDIO_FILE;
	if (!mmd->afd.afhi.chunks_total || !mmd->afd.afhi.seconds_total)
		goto out;
	promille = (1000 * mmd->current_chunk) / mmd->afd.afhi.chunks_total;
	if (backwards)
		promille -= 1000 * i / mmd->afd.afhi.seconds_total;
	else
		promille += 1000 * i / mmd->afd.afhi.seconds_total;
	if (promille < 0)
		promille = 0;
	if (promille >  1000) {
		mmd->new_vss_status_flags |= VSS_NEXT;
		goto out;
	}
	mmd->repos_request = (mmd->afd.afhi.chunks_total * promille) / 1000;
	mmd->new_vss_status_flags |= VSS_REPOS;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	mmd->events++;
	ret = 1;
out:
	mutex_unlock(mmd_mutex);
	return ret;
}

/* jmp */
int com_jmp(__a_unused struct rc4_context *rc4c, int argc, char * const * argv)
{
	long unsigned int i;
	int ret;

	if (argc != 2)
		return -E_COMMAND_SYNTAX;
	if (sscanf(argv[1], "%lu", &i) <= 0)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	ret = -E_NO_AUDIO_FILE;
	if (!mmd->afd.afhi.chunks_total)
		goto out;
	if (i > 100)
		i = 100;
	PARA_INFO_LOG("jumping to %lu%%\n", i);
	mmd->repos_request = (mmd->afd.afhi.chunks_total * i + 50)/ 100;
	PARA_INFO_LOG("sent: %lu,  offset before jmp: %lu\n",
		mmd->chunks_sent, mmd->offset);
	mmd->new_vss_status_flags |= VSS_REPOS;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	ret = 1;
	mmd->events++;
out:
	mutex_unlock(mmd_mutex);
	return ret;
}

/*
 * check if perms are sufficient to exec a command having perms cmd_perms.
 * Returns 0 if perms are sufficient, -E_PERM otherwise.
 */
static int check_perms(unsigned int perms, struct server_command *cmd_ptr)
{
	PARA_DEBUG_LOG("checking permissions\n");
	return (cmd_ptr->perms & perms) < cmd_ptr->perms ? -E_PERM : 0;
}

/*
 * Parse first string from *cmd and lookup in table of valid commands.
 * On error, NULL is returned.
 */
static struct server_command *parse_cmd(const char *cmdstr)
{
	char buf[255];
	int n = 0;

	sscanf(cmdstr, "%200s%n", buf, &n);
	if (!n)
		return NULL;
	buf[n] = '\0';
	return get_cmd_ptr(buf, NULL);
}

static int read_command(struct rc4_context *rc4c, char **result)
{
	int ret;
	char buf[4096];
	char *command = NULL;

	for (;;) {
		size_t numbytes;
		char *p;

		ret = rc4_recv_buffer(rc4c, buf, sizeof(buf));
		if (ret < 0)
			goto out;
		if (!ret)
			break;
		numbytes = ret;
		ret = -E_COMMAND_SYNTAX;
		if (command && numbytes + strlen(command) > MAX_COMMAND_LEN) /* DOS */
			goto out;
		command = para_strcat(command, buf);
		p = strstr(command, EOC_MSG);
		if (p) {
			*p = '\0';
			break;
		}
	}
	ret = command? 1 : -E_COMMAND_SYNTAX;
out:
	if (ret < 0)
		free(command);
	else
		*result = command;
	return ret;

}

static void reset_signals(void)
{
	para_sigaction(SIGCHLD, SIG_IGN);
	para_sigaction(SIGINT, SIG_DFL);
	para_sigaction(SIGTERM, SIG_DFL);
	para_sigaction(SIGHUP, SIG_DFL);
}

/**
 * Perform user authentication and execute a command.
 *
 * \param fd The file descriptor to send output to.
 * \param peername Identifies the connecting peer.
 *
 * Whenever para_server accepts an incoming tcp connection on
 * the port it listens on, it forks and the resulting child
 * calls this function.
 *
 * An RSA-based challenge/response is used to authenticate
 * the peer. It that authentication succeeds, a random RC4
 * session key is generated and sent back to the peer,
 * encrypted with its RSA public key.  From this point on,
 * all transfers are crypted with this session key.
 *
 * Next it is checked if the peer supplied  a valid server command or a command
 * for the audio file selector.  If yes, and if the user has sufficient
 * permissions to execute that command, the function calls the corresponding
 * command handler which does argument checking and further processing.
 *
 * In order to cope with a DOS attacks, a timeout is set up
 * which terminates the function if the connection was not
 * authenticated when the timeout expires.
 *
 * \sa alarm(2), rc4(3), crypt.c, crypt.h
 */
__noreturn void handle_connect(int fd, const char *peername)
{
	int ret, argc;
	char buf[4096];
	unsigned char rand_buf[CHALLENGE_SIZE + 2 * RC4_KEY_LEN];
	unsigned char challenge_sha1[HASH_SIZE];
	struct user *u;
	struct server_command *cmd = NULL;
	char **argv = NULL;
	char *p, *command = NULL;
	size_t numbytes;
	struct rc4_context rc4c = {.fd = fd};

	reset_signals();
	/* we need a blocking fd here as recv() might return EAGAIN otherwise. */
	ret = mark_fd_blocking(fd);
	if (ret < 0)
		goto err_out;
	/* send Welcome message */
	ret = send_va_buffer(fd, "This is para_server, version "
		PACKAGE_VERSION  ".\n" );
	if (ret < 0)
		goto err_out;
	/* recv auth request line */
	ret = recv_buffer(fd, buf, sizeof(buf));
	if (ret < 0)
		goto err_out;
	if (ret < 10) {
		ret = -E_AUTH_REQUEST;
		goto err_out;
	}
	numbytes = ret;
	ret = -E_AUTH_REQUEST;
	if (strncmp(buf, AUTH_REQUEST_MSG, strlen(AUTH_REQUEST_MSG)))
		goto err_out;
	p = buf + strlen(AUTH_REQUEST_MSG);
	PARA_DEBUG_LOG("received auth request for user %s\n", p);
	ret = -E_BAD_USER;
	u = lookup_user(p);
	if (u) {
		get_random_bytes_or_die(rand_buf, sizeof(rand_buf));
		ret = para_encrypt_buffer(u->rsa, rand_buf, sizeof(rand_buf),
			(unsigned char *)buf);
		if (ret < 0)
			goto err_out;
		numbytes = ret;
	} else {
		/*
		 * We don't want to reveal our user names, so we send a
		 * challenge to the client even if the user does not exist, and
		 * fail the authentication later.
		 */
		numbytes = 256;
		get_random_bytes_or_die((unsigned char *)buf, numbytes);
	}
	PARA_DEBUG_LOG("sending %u byte challenge + rc4 keys (%zu bytes)\n",
		CHALLENGE_SIZE, numbytes);
	ret = send_bin_buffer(fd, buf, numbytes);
	if (ret < 0)
		goto net_err;
	/* recv challenge response */
	ret = recv_bin_buffer(fd, buf, HASH_SIZE);
	if (ret < 0)
		goto net_err;
	numbytes = ret;
	PARA_DEBUG_LOG("received %d bytes challenge response\n", ret);
	ret = -E_BAD_USER;
	if (!u)
		goto net_err;
	/*
	 * The correct response is the sha1 of the first CHALLENGE_SIZE bytes
	 * of the random data.
	 */
	ret = -E_BAD_AUTH;
	if (numbytes != HASH_SIZE)
		goto net_err;
	sha1_hash((char *)rand_buf, CHALLENGE_SIZE, challenge_sha1);
	if (memcmp(challenge_sha1, buf, HASH_SIZE))
		goto net_err;
	/* auth successful */
	alarm(0);
	PARA_INFO_LOG("good auth for %s\n", u->name);
	/* init rc4 keys with the second part of the random buffer */
	RC4_set_key(&rc4c.recv_key, RC4_KEY_LEN, rand_buf + CHALLENGE_SIZE);
	RC4_set_key(&rc4c.send_key, RC4_KEY_LEN, rand_buf + CHALLENGE_SIZE
		+ RC4_KEY_LEN);
	ret = rc4_send_buffer(&rc4c, PROCEED_MSG);
	if (ret < 0)
		goto net_err;
	ret = read_command(&rc4c, &command);
	if (ret == -E_COMMAND_SYNTAX)
		goto err_out;
	if (ret < 0)
		goto net_err;
	ret = -E_BAD_CMD;
	cmd = parse_cmd(command);
	if (!cmd)
		goto err_out;
	/* valid command, check permissions */
	ret = check_perms(u->perms, cmd);
	if (ret < 0)
		goto err_out;
	/* valid command and sufficient perms */
	ret = create_argv(command, "\n", &argv);
	if (ret < 0)
		goto err_out;
	argc = ret;
	PARA_NOTICE_LOG("calling com_%s() for %s@%s\n", cmd->name, u->name,
			peername);
	ret = cmd->handler(&rc4c, argc, argv);
	free_argv(argv);
	mutex_lock(mmd_mutex);
	mmd->num_commands++;
	mutex_unlock(mmd_mutex);
	if (ret >= 0)
		goto out;
err_out:
	rc4_send_va_buffer(&rc4c, "%s\n", para_strerror(-ret));
net_err:
	PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
out:
	free(command);
	mutex_lock(mmd_mutex);
	if (cmd && (cmd->perms & AFS_WRITE) && ret >= 0)
		mmd->events++;
	mmd->active_connections--;
	mutex_unlock(mmd_mutex);
	exit(ret < 0? EXIT_FAILURE : EXIT_SUCCESS);
}
