/*
 * Copyright (C) 2005-2010 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file write.c Paraslash's standalone wav/raw player. */

#include <regex.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdbool.h>

#include "para.h"
#include "string.h"
#include "write.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "stdin.h"
#include "buffer_tree.h"
#include "write.h"
#include "write_common.h"
#include "fd.h"
#include "error.h"

INIT_WRITE_ERRLISTS;

enum check_wav_state {
	CWS_NEED_HEADER,
	CWS_HAVE_HEADER,
	CWS_NO_HEADER,
};

struct check_wav_task {
	int state;
	/** Number of channels specified in wav header given by \a buf. */
	unsigned channels;
	/** Sample rate specified in wav header given by \a buf. */
	unsigned samplerate;
	/** The task structure used by the scheduler. */
	struct task task;
	struct btr_node *btrn;
	size_t min_iqs;
};

static struct write_args_info conf;

static struct stdin_task sit;

/** Length of a standard wav header. */
#define WAV_HEADER_LEN 44

/**
 * Test if audio buffer contains a valid wave header.
 *
 * \return If not, return -E_NO_WAV_HEADER, otherwise, return zero. If
 * there is less than WAV_HEADER_LEN bytes available, return one.
 */
static void check_wav_pre_select(struct sched *s, struct task *t)
{
	struct check_wav_task *cwt = container_of(t, struct check_wav_task, task);
	int ret;

	ret = btr_node_status(cwt->btrn, cwt->min_iqs, BTR_NT_INTERNAL);
	if (ret != 0)
		sched_min_delay(s);
}

static int check_wav_exec(struct btr_node *btrn, const char *cmd, char **result)
{
	struct check_wav_task *cwt = btr_context(btrn);


	if (!strcmp(cmd, "samplerate")) {
		if (cwt->state != CWS_HAVE_HEADER)
			return -E_BTR_NAVAIL;
		*result = make_message("%d", cwt->samplerate);
		return 1;
	}
	if (!strcmp(cmd, "channels")) {
		if (cwt->state != CWS_HAVE_HEADER)
			return -E_BTR_NAVAIL;
		*result = make_message("%d", cwt->channels);
		return 1;
	}
	return -ERRNO_TO_PARA_ERROR(ENOTSUP);
}

static void check_wav_post_select(__a_unused struct sched *s, struct task *t)
{
	struct check_wav_task *cwt = container_of(t, struct check_wav_task, task);
	struct btr_node *btrn = cwt->btrn;
	unsigned char *a;
	size_t sz;
	int ret;

	t->error = 0;
	ret = btr_node_status(btrn, cwt->min_iqs, BTR_NT_INTERNAL);
	if (ret <= 0)
		goto out;
	if (cwt->state != CWS_NEED_HEADER)
		goto pushdown;
	btr_merge(btrn, cwt->min_iqs);
	sz = btr_next_buffer(btrn, (char **)&a);
	if (sz < cwt->min_iqs) /* file size less than WAV_HEADER_SIZE */
		goto pushdown;
	cwt->min_iqs = 0;
	cwt->channels = 2;
	cwt->samplerate = 44100;
	if (a[0] != 'R' || a[1] != 'I' || a[2] != 'F' || a[3] != 'F') {
		PARA_NOTICE_LOG("wav header not found\n");
		cwt->state = CWS_NO_HEADER;
		sprintf(t->status, "check wav: no header");
		goto out;
	}
	PARA_INFO_LOG("found wav header\n");
	cwt->state = CWS_HAVE_HEADER;
	sprintf(t->status, "check wav: have header");
	cwt->channels = (unsigned) a[22];
	cwt->samplerate = a[24] + (a[25] << 8) + (a[26] << 16) + (a[27] << 24);
	PARA_INFO_LOG("channels: %d, sample rate: %d\n", cwt->channels, cwt->samplerate);
	btr_consume(btrn, WAV_HEADER_LEN);
pushdown:
	btr_pushdown(btrn);
out:
	t->error = ret;
	if (ret < 0)
		btr_remove_node(btrn);
}

static int loglevel;
INIT_STDERR_LOGGING(loglevel)

__noreturn static void print_help_and_die(void)
{
	int d = conf.detailed_help_given;
	const char **p = d? write_args_info_detailed_help
		: write_args_info_help;

	printf_or_die("%s\n\n", WRITE_CMDLINE_PARSER_PACKAGE "-"
		WRITE_CMDLINE_PARSER_VERSION);
	printf_or_die("%s\n\n", write_args_info_usage);
	for (; *p; p++)
		printf_or_die("%s\n", *p);
	print_writer_helps(d);
	exit(0);
}

static int main_btr(struct sched *s)
{
	int i, ret;
	struct check_wav_task _cwt, *cwt = &_cwt;
	struct writer_node *wns;

	loglevel = get_loglevel_by_name(conf.loglevel_arg);
	sit.btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "stdin"));
	stdin_set_defaults(&sit);
	register_task(&sit.task);

	cwt->state = CWS_NEED_HEADER;
	cwt->min_iqs = WAV_HEADER_LEN;
	cwt->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "check_wav", .parent = sit.btrn,
		.handler = check_wav_exec, .context = cwt));
	sprintf(cwt->task.status, "check_wav");
	cwt->task.pre_select = check_wav_pre_select;
	cwt->task.post_select = check_wav_post_select;
	cwt->task.error = 0;
	register_task(&cwt->task);

	ret = -E_WRITE_SYNTAX;
	if (!conf.writer_given) {
		i = 0;
		wns = para_calloc(sizeof(*wns));
		ret = setup_writer_node(NULL, cwt->btrn, wns);
		if (ret < 0)
			goto out;
		i = 1;
	} else {
		wns = para_calloc(conf.writer_given * sizeof(*wns));
		for (i = 0; i < conf.writer_given; i++) {
			ret = setup_writer_node(conf.writer_arg[i],
				cwt->btrn, wns + i);
			if (ret < 0)
				goto out;
		}
	}

	s->default_timeout.tv_sec = 10;
	s->default_timeout.tv_usec = 50000;
	ret = schedule(s);
out:
	for (i--; i >= 0; i--) {
		struct writer_node *wn = wns + i;
		struct writer *w = writers + wn->writer_num;

		w->close(wn);
		btr_free_node(wn->btrn);
		free(wn->conf);
	}
	free(wns);
	btr_free_node(cwt->btrn);
	return ret;
}

/**
 * Para_write's main function.
 *
 * \param argc The usual argument counter.
 * \param argv The usual argument vector.
 *
 * It sets up and starts the tasks and the buffer tree nodes determined by
 * command line options.
 *
 * \return \p EXIT_SUCCESS or EXIT_FAILURE
 */
int main(int argc, char *argv[])
{
	int ret = -E_WRITE_SYNTAX;
	static struct sched s;

	writer_init();
	write_cmdline_parser(argc, argv, &conf);
	HANDLE_VERSION_FLAG("write", conf);
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();

	ret = main_btr(&s);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
	exit(EXIT_SUCCESS);
}
