/*
 * Copyright (C) 2006-2010 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file write.h Writer-related structures. */

/** The list of supported writers. */
enum writer_enum {WRITER_ENUM};

/**
 * Describes one running instance of a writer.
 */
struct writer_node {
	/** The number of this writer. */
	int writer_num;
	/** Writer-specific data. */
	void *private_data;
	/** The writer-specific configuration of this node. */
	void *conf;
	struct btr_node *btrn;
	struct task task;
	size_t min_iqs;
};

/** Describes one supported writer. */
struct writer {
	/**
	 * The init function of the writer.
	 *
	 * It must fill in all other function pointers of the given
	 * writer structure.
	 */
	void (*init)(struct writer *w);
	/**
	 * The command line parser of the writer.
	 *
	 * It should check whether the command line options given by \a options are
	 * valid.  On success, it should return a pointer to the writer-specific
	 * configuration data determined by \a options.  Note that this might be called
	 * more than once with different values of \a options.
	 */
	void *(*parse_config)(const char *options);
	void (*free_config)(void *conf);
	/**
	 * Open one instance of this writer.
	 *
	 * This function should perform any work necessary to write the incoming
	 * stream. To this aim, it may allocate its private data structure and store
	 * a pointer to that structure via the given writer_node parameter.
	 */
	int (*open)(struct writer_node *);
	/**
	 * Prepare the fd sets for select.
	 *
	 * This is called from scheduler. It may use the sched pointer to add
	 * any file descriptors or to decrease the select timeout.
	 */
	void (*pre_select)(struct sched *s, struct task *t);
	/**
	 * Write audio data.
	 *
	 * Called from the post_select function of the writer node's task.
	 */
	void (*post_select)(struct sched *s, struct task *t);
	/**
	 * Close one instance of the writer.
	 *
	 * This function is assumed to succeed.
	 */
	void (*close)(struct writer_node *);
	/**
	 * Shutdown the writer
	 *
	 * This is a optional function pointer used for cleaning up.
	 */
	void (*shutdown)(struct writer_node *);
	struct ggo_help help;
	btr_command_handler execute;
};

/** Loop over each supported writer. */
#define FOR_EACH_WRITER(i) for (i = 0; i < NUM_SUPPORTED_WRITERS; i++)

/** Declare the init functions of all supported writers. */
DECLARE_WRITER_INITS;

/** Array containing the name of each writer. */
extern const char *writer_names[];

/** The writer structure for each supported writer. */
extern struct writer writers[NUM_SUPPORTED_WRITERS];
